# TP Hachage



 
# Partie 1 - Réaliser une fonction de hachage minimale parfaite 

On travaillera sur `tp_2_miso_mphf.py`.

Tout d'abord, voyons ensemble l'idée de la construction de la MPHF.
Puis suivez les commentaires dans le code à compléter.
Vous devez 
- finaliser la fonction pour construire la MPHF (`construction_mphf`), 
- écrire la fonction pour obtenir le hash d'un élément par la MPHF (`get_hash_mphf`), 
- écrire la fonction pour créer une table de hachage avec cette mphf (`create_hash_table`)

Ecrivez vos réponses et commentaires dans ce document.
Puis décommenter `compare_taille` à la fin et expliquer les résultats.

###Explication des courble sur l'image  obtenu apres execution du code:
Sur le graphique, on compare la mémoire utilisée par deux structures : une table de hachage construite avec une MPHF (en bleu) et un dictionnaire Python classique (en orange), en fonction du nombre de k-mers qu’on y stocke.

La table avec MPHF :
La courbe monte doucement et de façon régulière. Cela montre que la MPHF utilise la mémoire de manière très efficace : elle n’alloue que ce qui est strictement nécessaire. Elle ne laisse pas d’espace vide inutile et reste donc très compacte, ce qui est parfait quand on connaît à l’avance tous les k-mers qu’on veut stocker.

Le dictionnaire :
La courbe est beaucoup moins fluide. Elle monte par paliers, car le dictionnaire doit régulièrement agrandir sa structure interne pour pouvoir continuer à insérer des éléments. À chaque palier, il réserve plus d’espace que nécessaire, ce qui entraîne un gaspillage de mémoire. Et plus le nombre de k-mers augmente, plus ça devient lourd.
 En résumé :
La MPHF est clairement plus avantageuse en termes de mémoire, surtout quand on travaille avec de très grands ensembles de k-mers, comme c’est souvent le cas en bioinformatique. Le dictionnaire, lui, reste pratique et facile à utiliser, mais il devient vite inefficace quand on monte en volume.

Bonus : faites varier `nb_niveaux` et `gamma`, voyez quelle influence ils peuvent avoir.
On constate qu'augmenter gamma ou nb_niveaux ne change pas la taille finale de la structure (elle reste fixe car on stocke une paire par k-mer),
mais cela ralentit la construction de la MPHF. 
Autrement dit, ces paramètres font travailler le code un peu plus pour mieux gérer les collisions, ce qui allonge le temps de calcul sans affecter la taille finale.

# Partie 2 - Analyse de performance de dictionnaires en Python

On travaillera sur `tp_2_miso_dict.py`

A la fin de cette partie, en lançant `python tp_2_miso_dict.py` on doit obtenir 4 graphiques.

## Introduction

Nous allons analyser la performance de dictionnaires `dict` en Python en fonction de différents facteurs : que le facteur de charge (loading factor), 
le temps d'insertion et la taille de la mémoire occupée. Vous pourrez utiliser matplotlib et numpy pour visualiser les résultats.

##  Étude du facteur de charge

Fonction `experiment_load_factor`:

0. La fonction `experiment_load_factor` doit être définie pour prendre en entrée une liste de facteurs de charge (`load_factors`) et renvoyer les temps 
- d'insertion de clefs, 
- les nombres de réallocations de mémoire (quand le dictionnaire est ré-écrit en mémoire pour agrandir sa taille) 
- et les tailles de mémoire occupées par le dictionnaire pour chaque facteur de charge. 

1. Initialisation des listes pour stocker les résultats.
Initialiser dans la fonction `experiment_load_factor` des listes `insertion_times`, `num_resizes` et `sizes` seront utilisées pour stocker les trois résultats décrits au dessus.

2. La fonction `experiment_load_factor` doit boucler sur chaque facteur de charge de la liste en entrée `load_factors`. Créer un dictionnaire d'abord vide à chaque fois.

3. Initialiser les variables `num_elements`, `start_time`, `num_resize` et `last_size` pour respectivement mesurer  le nombre d'éléments, le temps d'insertion et le nombre de réallocations de mémoire.

4. Insérer les éléments dans le dictionnaire et mesurer le temps d'insertion pour chaque élément, vérifier le nombre de réallocations mémoire (utiliser par exemple `time.time()` et `sys.getsizeof()` pour mesurer le temps et la taille du dictionnaire avant et après chaque insertion).

5. Pour un facteur de charge donné, stocker les résultats dans les listes `insertion_times`, `num_resizes` et `sizes`.

## Deuxième étude

6. A quoi sert la fonction `experiment_longest` ?

## Visualisation des résultats

7. Créez quatre graphiques au format png ou pdf : 

- Un graphique du temps d'insertion en fonction du facteur de charge (obtenu question 5)
- Un graphique du nombre de réallocations de mémoire en fonction du facteur de charge (obtenu question 5)
- Un graphique de la taille de mémoire occupée en fonction du nombre d'éléments (obtenu question 5)
- Un histogramme des fréquences des temps d'insertions discrétisés (code fourni, remplacer la liste vide par la bonne entrée)

10. Commentez vos résultats.

 -Pour le graphique du temps d'insertion en fonction du facteur de charge ,on constate que l’insertion dans un dictionnaire Python demeure rapide même lorsque le facteur de charge est proche de 1.0.
Les sauts  traduisent  les réallocations internes et la gestion mémoire dynamique de Python, qui peuvent se produire à des moments différents d’une exécution à l’autre

-Pour le graphique du nombre de réallocations de mémoire en fonction du facteur de charge , on constate que plus on insère d’éléments (donc plus le facteur de charge est élevé), 
 plus le dictionnaire Python se réalloue en mémoire, jusqu’à atteindre un plateau autour de 12 réallocations.
 Python effectue ces réallocations par paliers, ce qui explique pourquoi le nombre total de «resizes» reste finalement assez limité malgré un grand nombre d’insertions.
 
-Pour le graphique de la taille de mémoire occupée en fonction du nombre d'éléments ,On voit que la mémoire occupée par le dictionnaire augmente à mesure que le nombre d’éléments insérés croît, 
 avec des sauts plus ou moins marqués qui correspondent aux réallocations internes. Au-delà d’un certain seuil (environ 6000 éléments ici), la taille mémoire se stabilise, car le dictionnaire a suffisamment grandi pour accueillir la totalité des clés. 
 
-On constate sur l'histogramme des fréquences des temps d'insertions discrétisés que La majorité des insertions se font en un temps quasi nul , tandis qu’une poignée d’insertions plus longues apparaissent dans la seconde barre, 
probablement lors des réallocations.
Cela souligne qu’en Python, la plupart des insertions sont très rapides, mais certaines opérations de réallocation peuvent entraîner un temps d’insertion plus élevé.